@@ -31,11 +31,23 @@ void sort_as_appropriate(void);
 #ifdef HAVE_LIBID3
 static int get_id3_field(ID3Tag *tag, ID3_FrameID frameid, char *buf, int len);
 #endif
+#ifdef HAVE_LIBID3TAG
+static id3_latin1_t *get_id3tag_field(struct id3_tag *tag, char *id);
+#endif
 #ifdef HAVE_NFTW
 static int add_directory_callback(const char *name, const struct stat *sb,
 				  int flag, struct FTW *ftw);
 #endif
 static void kill_child(int signal);
+#ifdef USEGTK2
+static gboolean iter_from_position(gint item, GtkTreeIter *iter);
+#endif
+static void freeze_list();
+static void thaw_list();
+static void select_list_row(gint item);
+static void clear_list();
+static void remove_list_row(gint item);
+static void insert_list_row(songinfo *sinfo, gint item);
 
 /* Globals. */
 pid_t childpid = 0;
@@ -47,6 +59,108 @@ gint last_played = -1,
   last_row_hit;
 fileaction *last_action = NULL;
 
+#ifdef USEGTK2
+/* Get a GtkTreeIter pointing at a given row in the list. */
+static gboolean iter_from_position(gint item, GtkTreeIter *iter) {
+  return gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(filestore), iter,
+				       NULL, item);
+}
+#endif
+
+/* Freeze the song list. */
+static void freeze_list() {
+#ifdef USEGTK2
+  /* TreeViews don't appear to support freezing/thawing. */
+#else
+  gtk_clist_freeze(GTK_CLIST(filelist));
+#endif
+}
+
+/* Thaw the song list. */
+static void thaw_list() {
+#ifndef USEGTK2
+  gtk_clist_thaw(GTK_CLIST(filelist));
+#endif
+}
+
+/* Select a given row in the list. It's assumed that this calls the
+   item-selected callback. */
+static void select_list_row(gint item) {
+#ifdef USEGTK2
+  GtkTreeIter iter;
+  GtkTreeSelection *selection;
+
+  iter_from_position(item, &iter);
+  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(filelist));
+  gtk_tree_selection_select_iter(selection, &iter);
+#else
+  gtk_clist_select_row(GTK_CLIST(filelist), item, 0);
+#endif
+}
+
+/* Clear the list. */
+static void clear_list() {
+#ifdef USEGTK2
+  GtkTreeIter iter;
+
+  /* Destroy all the list's songinfo structures. */
+  if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(filestore), &iter)) {
+    do {
+      gpointer sinfo;
+
+      gtk_tree_model_get(GTK_TREE_MODEL(filestore), &iter,
+			 SONGINFO_COLUMN, &sinfo,
+			 -1);
+      destroy_songinfo(sinfo);
+    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(filestore), &iter));
+  }
+  
+  gtk_list_store_clear(GTK_LIST_STORE(filestore));
+#else
+  gtk_clist_clear(GTK_CLIST(filelist));
+#endif
+}
+
+/* Remove a row from the list. */
+static void remove_list_row(gint item) {
+#ifdef USEGTK2
+  GtkTreeIter iter;
+  gpointer sinfo;
+
+  iter_from_position(item, &iter);
+  gtk_tree_model_get(GTK_TREE_MODEL(filestore), &iter,
+		     SONGINFO_COLUMN, &sinfo,
+		     -1);
+  destroy_songinfo(sinfo);
+  gtk_list_store_remove(GTK_LIST_STORE(filestore), &iter);
+#else
+  gtk_clist_remove(GTK_CLIST(filelist), item);
+#endif
+}
+
+/* Insert a row in the list. */
+static void insert_list_row(songinfo *sinfo, gint item) {
+#ifdef USEGTK2
+  GtkTreeIter iter;
+
+  gtk_list_store_insert(GTK_LIST_STORE(filestore), &iter, item);
+  gtk_list_store_set(GTK_LIST_STORE(filestore), &iter,
+		     SONGINFO_COLUMN, sinfo,
+		     TITLE_COLUMN, sinfo->display_name,
+		     -1);
+  /* Note: in the CList version, we add destroy_songinfo as a hook
+     with the data. In the Tree version, we call it when rows are
+     removed. */
+#else
+  gchar *row[1];
+  
+  row[0] = sinfo->display_name;
+  gtk_clist_insert(GTK_CLIST(filelist), item, row);
+  gtk_clist_set_row_data_full(GTK_CLIST(filelist), item, sinfo,
+			      destroy_songinfo);
+#endif
+}
+
 /* Send a signal to the child process (or process group). */
 static void kill_child(int signal) {
   kill(last_action->setgroup ? -childpid : childpid, signal);
@@ -123,6 +237,7 @@ void start_playing(gint num) {
   name = get_songinfo(num)->name;
   last_action = get_action(name);
   if (!last_action) {
+    playing = 0;
     status(_("Unable to guess type."));
     return;
   }
@@ -224,14 +339,13 @@ void pick_prev() {
 	newitem = listcount - 1;
       }
     }
-    gtk_clist_select_row(GTK_CLIST(filelist), newitem, 0);
+    select_list_row(newitem);
     /* ... which calls handle_list_select. */
   } else {
     status(_("No items to pick from!"));
   }
 }
 
-
 /* Pick and play a random song. */
 void pick_next() {
   int newitem, first, i;
@@ -243,6 +357,7 @@ void pick_next() {
 	newitem = 0;
         if (mode_one_time) {
           stop_playing();
+	  status(_("All items played."));
           return;
         }
       }
@@ -271,7 +386,7 @@ void pick_next() {
 	get_songinfo(newitem)->previous = last_played;
       }
     }
-    gtk_clist_select_row(GTK_CLIST(filelist), newitem, 0);
+    select_list_row(newitem);
     /* ... which calls handle_list_select. */
   } else {
     status(_("No items to pick from!"));
@@ -296,6 +411,23 @@ static int get_id3_field(ID3Tag *tag, ID3_FrameID frameid,
 }
 #endif
 
+#ifdef HAVE_LIBID3TAG
+static id3_latin1_t *get_id3tag_field(struct id3_tag *tag, char *id) {
+  struct id3_frame const *frame;
+  union id3_field const *field;
+  id3_ucs4_t const *ucs4;
+
+  frame = id3_tag_findframe(tag, id, 0);
+  if (!frame) return NULL;
+  field = &frame->fields[1];
+  if (id3_field_getnstrings(field) == 0) return NULL;
+  ucs4 = id3_field_getstrings(field, 0);
+  if (!ucs4) return NULL;
+  /* FIXME GTK2 expects UTF8. */
+  return id3_ucs4_latin1duplicate(ucs4);
+}
+#endif
+
 /* Return the "Title - Artist" form of a song's title from its ID3
    tag; return NULL if one can't be found. */
 gchar *read_mp3_tag (const char *filename) {
@@ -324,6 +456,32 @@ gchar *read_mp3_tag (const char *filename) {
  out:
   ID3Tag_Delete(tag);
   return trackname;
+#else
+#ifdef HAVE_LIBID3TAG
+  /* Use libid3tag to read the tag. */
+  struct id3_file *f;
+  struct id3_tag *t;
+  id3_latin1_t *title = NULL, *artist = NULL;
+
+  f = id3_file_open(filename, ID3_FILE_MODE_READONLY);
+  if (!f) return NULL;
+
+  t = id3_file_tag(f);
+  if (!t) goto out;
+
+  title = get_id3tag_field(t, ID3_FRAME_TITLE);
+  if (!title) goto out;
+  artist = get_id3tag_field(t, ID3_FRAME_ARTIST);
+  if (!artist) goto out;
+
+  trackname = malloc(strlen(artist) + 3 + strlen(title) + 1);
+  sprintf(trackname, "%s - %s", artist, title);
+
+ out: 
+  if (title) free(title);
+  if (artist) free(artist);
+  id3_file_close(f);
+  return trackname;
 #else
   /* Use our simple version. Routine to read the ID3 tag from an MP3
      file, originally supplied by Craig Knudsen. */
@@ -347,6 +505,7 @@ gchar *read_mp3_tag (const char *filename) {
   }
   close(fd);
   return trackname;
+#endif /* HAVE_LIBID3TAG */
 #endif /* HAVE_LIBID3 */
 }
 
@@ -473,7 +632,7 @@ static int add_directory_callback(const char *name, const struct stat *sb,
    playlist. */
 void add_directory(const gchar *name) {
 #ifdef HAVE_NFTW
-  gtk_clist_freeze(GTK_CLIST(filelist));
+  freeze_list();
 
   /* We use an arbitrary maximum depth of 100, since nftw (at least in
      the glibc implementation) doesn't have an option for "unlimited
@@ -482,7 +641,7 @@ void add_directory(const gchar *name) {
   nftw(name, add_directory_callback, 100, FTW_DEPTH);
 
   sort_as_appropriate();
-  gtk_clist_thaw(GTK_CLIST(filelist));
+  thaw_list();
 #else
   /* We don't have nftw; just don't do anything (for now). */
 #endif
@@ -491,7 +650,7 @@ void add_directory(const gchar *name) {
 /* Clear the playlist. */
 void clear_playlist() {
   stop_playing();
-  gtk_clist_clear(GTK_CLIST(filelist));
+  clear_list();
   last_played = -1;
   update_window();
 }
@@ -563,16 +722,16 @@ void sortplaylist(int (*func)(const void *a, const void *b),
   for (i=0; i<lc; i++) sl[i] = copy_songinfo(get_songinfo(i));
   qsort(sl, lc, sizeof(songinfo *), func);
   
-  gtk_clist_freeze(GTK_CLIST(filelist));
+  freeze_list();
   clear_playlist();
   for (i=0; i<lc; i++) insert_row(sl[i], i);
-  gtk_clist_thaw(GTK_CLIST(filelist));
+  thaw_list();
 
   free(sl);
 }
 
 /* Write a playlist. Returns 1 on error, 0 on success. */
-int write_playlist(gchar *name) {
+int write_playlist(const gchar *name) {
   int i;
   FILE *f = fopen(name, "w");
 
@@ -589,8 +748,7 @@ void read_playlist(gchar *name, int use_cwd) {
   char line[STRBUFSIZE], fullname[STRBUFSIZE], dirname[STRBUFSIZE], 
     *slash, *filename;
 
-  /* Prevent the list from being updated while we load the list. */
-  gtk_clist_freeze(GTK_CLIST(filelist));
+  freeze_list();
 
   f = fopen(name, "r");
   if (!f) {
@@ -652,8 +810,8 @@ void read_playlist(gchar *name, int use_cwd) {
     gtk_window_set_title(GTK_WINDOW(window), t);
     free(t);
   }
-  
-  gtk_clist_thaw(GTK_CLIST(filelist));
+
+  thaw_list();
 }
 
 /* Copy a songinfo structure. */
@@ -699,36 +857,43 @@ void delete_row(gint location) {
     if (s->previous > location) --s->previous;
   }
 
-  gtk_clist_remove(GTK_CLIST(filelist), location);
+  remove_list_row(location);
 }
 
 /* Insert a row into the list based on a filled-in songinfo structure. 
    Call as insert_row(X, listcount); to append to the end. */
 void insert_row(songinfo *sinfo, gint location) {
-  gchar *row[1];
-  
-  row[0] = sinfo->display_name;
-  gtk_clist_insert(GTK_CLIST(filelist), location, row);
-  gtk_clist_set_row_data_full(GTK_CLIST(filelist), location, sinfo,
-			      destroy_songinfo);
+  insert_list_row(sinfo, location);
   listcount++;
 }
 
 /* Get a pointer to a song's songinfo. */
 songinfo *get_songinfo(gint number) {
+#ifdef USEGTK2
+  GtkTreeIter iter;
+  gpointer sinfo;
+
+  iter_from_position(number, &iter);
+  gtk_tree_model_get(GTK_TREE_MODEL(filestore), &iter,
+		     SONGINFO_COLUMN, &sinfo,
+		     -1);
+  return (songinfo *)sinfo;
+#else
   return (songinfo *)gtk_clist_get_row_data(GTK_CLIST(filelist), number);
+#endif
 }
 
 /* Move a row from one location to another within the list. */
 void move_row(gint src, gint dest) {
   songinfo *new;
 
-  gtk_clist_freeze(GTK_CLIST(filelist));
+  freeze_list();
 
   new = copy_songinfo(get_songinfo(src));
-  gtk_clist_remove(GTK_CLIST(filelist), src);
+  remove_list_row(src);
   insert_row(new, dest);
-  gtk_clist_thaw(GTK_CLIST(filelist));
+
+  thaw_list();
 }
 
 /* Save the persistent playlist. */
