@@ -26,9 +26,12 @@
 #include "xhippo.h"
 
 /* Prototypes for static functions. */
-static GtkWidget *add_button(gchar *label, gchar *tip, GtkSignalFunc callback,
-			     GtkWidget *container, gchar *name, gint accel);
-static GtkWidget *add_menu_item(gchar *label, GtkSignalFunc callback,
+typedef void (*ButtonFunc)(GtkButton *, gpointer);
+static GtkWidget *add_button(const gchar *label, const gchar *tip,
+			     ButtonFunc callback, GtkWidget *container,
+			     const gchar *name, gint accel);
+typedef void (*MenuFunc)(GtkMenuItem *, gpointer);
+static GtkWidget *add_menu_item(const gchar *label, MenuFunc callback,
 				gint accel);
 static void add_menu_separator(void);
 
@@ -37,16 +40,20 @@ GtkTooltips *tooltips;
 GtkWidget *window, *restartbutton, *stopbutton, *buttons_box, 
   *all_box, *savebutton, *sortbutton, *addbutton,
   *filelist, *statusbar, *nextbutton, *prevbutton, *pausebutton, *loadbutton,
-  *minicheckbox, *randomcheckbox, *fileselector=NULL, *list_box,
+  *minicheckbox, *randomcheckbox, *fileselector = NULL, *list_box,
   *scrollbar, *upbutton, *downbutton, *deletebutton, *popupmenu,
-  *up_item, *down_item, *delete_item, *info_item, *eventbox;
+  *up_item, *down_item, *delete_item, *info_item, *eventbox,
+  *prefs_window = NULL, *prefs_all_box;
+#ifdef USEGTK2
+GtkListStore *filestore;
+#endif
 GtkAdjustment *vadj;
 GtkAccelGroup *accel_group;
 int sigchld_pipe[2];
 gchar startup_cwd[STRBUFSIZE];
 
 /* Show a message on the status bar */
-void status(char *message) {
+void status(const char *message) {
   gtk_statusbar_pop(GTK_STATUSBAR(statusbar), contextid);
   gtk_statusbar_push(GTK_STATUSBAR(statusbar), contextid, message);
 }
@@ -75,10 +82,28 @@ void update_window() {
       }
     }
 
-    if (mode_scroll_catchup 
-	&& (gtk_clist_row_is_visible(GTK_CLIST(filelist), last_played)
-	    != GTK_VISIBILITY_FULL)) {
-      gtk_clist_moveto(GTK_CLIST(filelist), last_played, 0, 0.5, 0.0);
+    if (mode_scroll_catchup && GTK_WIDGET_REALIZED(filelist)) {
+#ifdef USEGTK2
+      GdkRectangle row, vis;
+      GtkTreePath *path = gtk_tree_path_new();
+      gtk_tree_path_append_index(path, last_played);
+
+      gtk_tree_view_get_cell_area(GTK_TREE_VIEW(filelist), path, NULL, &row);
+      gtk_tree_view_widget_to_tree_coords(GTK_TREE_VIEW(filelist),
+					  row.x, row.y, &row.x, &row.y);
+      gtk_tree_view_get_visible_rect(GTK_TREE_VIEW(filelist), &vis);
+
+      if ((row.y < vis.y)
+	  || (row.y + row.height) >= (vis.y + vis.height)) {
+	gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(filelist), path, NULL,
+				     TRUE, 0.5, 0.0);
+      }
+#else
+      if (gtk_clist_row_is_visible(GTK_CLIST(filelist), last_played)
+	  != GTK_VISIBILITY_FULL) {
+	gtk_clist_moveto(GTK_CLIST(filelist), last_played, 0, 0.5, 0.0);
+      }
+#endif
     }
   } else {
     snprintf(statbuf, STRBUFSIZE, _("Stopped."));
@@ -114,39 +139,6 @@ void handle_sigchld(int dummy) {
   write(sigchld_pipe[1], "!", 1);
 }
 
-#ifdef USEGNOMEMENUS
-
-static GnomeUIInfo file_menu[] = {
-  { GNOME_APP_UI_ITEM, "Load playlist...", NULL, handle_menu_load, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 'l', 0, NULL },
-  { GNOME_APP_UI_ITEM, "Restart", NULL, handle_restart, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 'r', 0, NULL },
-  { GNOME_APP_UI_ITEM, "Stop", NULL, handle_stop, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 's', 0, NULL },
-  { GNOME_APP_UI_ITEM, "Random", NULL, handle_next, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 'n', 0, NULL },
-  { GNOME_APP_UI_ITEM, "Quit", NULL, handle_destroy, NULL, NULL,
-    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_EXIT, 'Q',
-    GDK_SHIFT_MASK, NULL },
-  { GNOME_APP_UI_ENDOFINFO }
-};
-
-static GnomeUIInfo help_menu[] = {
-  { GNOME_APP_UI_ITEM, "About...", NULL, handle_about, NULL, NULL,
-    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_ABOUT, 0, 0, NULL },
-  { GNOME_APP_UI_ENDOFINFO }
-};
-
-static GnomeUIInfo main_menu[] = {
-  { GNOME_APP_UI_SUBTREE, ("File"), NULL, file_menu, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL },
-  { GNOME_APP_UI_SUBTREE, ("Help"), NULL, help_menu, NULL, NULL,
-    GNOME_APP_PIXMAP_NONE, NULL, 0, 0, NULL },
-  { GNOME_APP_UI_ENDOFINFO }
-};
-
-#endif /* USEGNOMEMENUS */
-
 static GtkTargetEntry target_table[] = {
   { "text/uri-list", 0, 0 },
 };
@@ -160,8 +152,9 @@ void print_version() {
 "see the file named COPYING.\n");
 }
 
-static GtkWidget *add_button(gchar *label, gchar *tip, GtkSignalFunc callback,
-			     GtkWidget *container, gchar *name, gint accel) {
+static GtkWidget *add_button(const gchar *label, const gchar *tip,
+			     ButtonFunc callback, GtkWidget *container,
+			     const gchar *name, gint accel) {
   GtkWidget *new;
 
   /* Create and bind the Load button. */
@@ -180,7 +173,7 @@ static GtkWidget *add_button(gchar *label, gchar *tip, GtkSignalFunc callback,
   return new;
 }
 
-static GtkWidget *add_menu_item(gchar *label, GtkSignalFunc callback,
+static GtkWidget *add_menu_item(const gchar *label, MenuFunc callback,
 				gint accel) {
   GtkWidget *item;
 
@@ -266,18 +259,19 @@ int main(int argc, char **argv) {
   /* In GNOME mode, window is actually a GNOME app, but the
      differences in use are very little. */
   window = gnome_app_new("xhippo", wintitlestring);
-#ifdef USEGNOMEMENUS
-  gnome_app_create_menus_with_data(GNOME_APP(window), main_menu, window);
-#endif
   gtk_widget_realize(window);
 #else
   /* Standard GTK initialisation. */
   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(window), wintitlestring);
 #endif
+
+#ifndef USEGTK2
+  /* With GTK1, mini mode only works when shrinking is enabled. With
+     GTK2, it only works when shrinking is disabled. Go figure. */
   gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);
-  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
-		     GTK_SIGNAL_FUNC(handle_delete_event), NULL);
+#endif
+
   gtk_signal_connect(GTK_OBJECT(window), "destroy",
 		     GTK_SIGNAL_FUNC(handle_destroy), NULL);
   gtk_signal_connect(GTK_OBJECT(window), "configure_event",
@@ -310,9 +304,13 @@ int main(int argc, char **argv) {
 
   /* Create the status bar. */
   statusbar = gtk_statusbar_new();
+#ifdef USEGTK2
+  gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), FALSE);
+#endif
   gtk_container_add (GTK_CONTAINER(eventbox), statusbar);
   contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "xhippo");
-  gtk_statusbar_push(GTK_STATUSBAR(statusbar), contextid, _("Welcome to xhippo!"));
+  gtk_statusbar_push(GTK_STATUSBAR(statusbar), contextid,
+		     _("Welcome to xhippo!"));
   gtk_widget_set_name(statusbar, "statusbar");
 
   gtk_widget_show(statusbar);
@@ -330,20 +328,20 @@ int main(int argc, char **argv) {
                              GDK_r, 0, GTK_ACCEL_VISIBLE);
 
   stopbutton = add_button(_("Stop"), _("Stop playing"), handle_stop,
-			   buttons_box, "stopbutton", GDK_KP_Divide);
+			  buttons_box, "stopbutton", GDK_KP_Divide);
   gtk_widget_add_accelerator(stopbutton, "clicked", accel_group,
                              GDK_s, 0, GTK_ACCEL_VISIBLE);
 
   pausebutton = add_button(_("Pause"), _("Pause playing"), handle_pause,
-			    buttons_box, "pausebutton", GDK_KP_Add);
+			   buttons_box, "pausebutton", GDK_KP_Add);
   gtk_widget_add_accelerator(pausebutton, "clicked", accel_group,
                              GDK_p, 0, GTK_ACCEL_VISIBLE);
 
   prevbutton = add_button(_("Prev"), _("Play the previous song"), handle_prev,
-			   buttons_box, "nextbutton", GDK_b);
+			  buttons_box, "nextbutton", GDK_b);
   
   nextbutton = add_button(_("Next"), _("Play the next song"), handle_next,
-			   buttons_box, "nextbutton", GDK_KP_Subtract);
+			  buttons_box, "nextbutton", GDK_KP_Subtract);
   gtk_widget_add_accelerator(nextbutton, "clicked", accel_group,
 			     GDK_KP_Multiply, 0, GTK_ACCEL_VISIBLE);
   gtk_widget_add_accelerator(nextbutton, "clicked", accel_group,
@@ -395,8 +393,14 @@ int main(int argc, char **argv) {
   add_menu_item(_("Sort playlist by mtime"), handle_menu_sort_mtime, GDK_t);
   add_menu_item(_("Clear playlist"), handle_menu_clear, GDK_c);
   add_menu_separator();
-  add_menu_item(_("Quit"), handle_destroy, GDK_q);
+#ifdef USEPREFS
+  add_menu_item(_("Preferences"), handle_menu_prefs, 0);
+#endif
+  add_menu_item(_("Quit"), handle_menu_quit, GDK_q);
   
+  /* Initialise the config system. */
+  init_config_commands();
+
   /* Read the config file for the first time. */
   read_config_files(0);
 
@@ -620,6 +624,7 @@ int main(int argc, char **argv) {
 
   /* Create the list. We have to do this after we've read the options,
      because we need to know the state of mode_left_scroll. */
+#ifndef USEGTK2
   filelist = gtk_clist_new(1);
 
   gtk_clist_set_selection_mode(GTK_CLIST(filelist), GTK_SELECTION_SINGLE);
@@ -630,11 +635,51 @@ int main(int argc, char **argv) {
   gtk_signal_connect(GTK_OBJECT(filelist), "select_row",
 		     GTK_SIGNAL_FUNC(handle_list_select), NULL);
   gtk_signal_connect(GTK_OBJECT(filelist), "button_press_event",
-		     GTK_SIGNAL_FUNC(handle_menu_popup), NULL);
+  		     GTK_SIGNAL_FUNC(handle_menu_popup), NULL);
+#else
+  {
+    GtkTreeViewColumn *column;
+    GtkCellRenderer *renderer;
+    GtkTreeSelection *select;
+    GtkTreeIter iter;
+
+    filestore = gtk_list_store_new(NUM_COLUMNS,
+				   G_TYPE_POINTER,
+				   G_TYPE_STRING);
+    filelist = gtk_tree_view_new_with_model(GTK_TREE_MODEL(filestore));
+    g_object_unref(G_OBJECT(filestore));
+
+    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(filelist), FALSE);
+    
+    renderer = gtk_cell_renderer_text_new();
+    column = gtk_tree_view_column_new_with_attributes("Title", renderer,
+						      "text", TITLE_COLUMN,
+						      NULL);
+    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
+    gtk_tree_view_append_column(GTK_TREE_VIEW(filelist), column);
+
+    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(filelist));
+    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
+    g_signal_connect(G_OBJECT(select), "changed",
+		     G_CALLBACK(handle_tree_select), NULL);
+    g_signal_connect(G_OBJECT(filelist), "button_press_event",
+		     G_CALLBACK(handle_menu_popup), NULL);
+  }
+
+  /* Since we add lots of items to the file list, we don't want it
+     requesting a really really tall size for itself. */
+  gtk_widget_set_size_request(GTK_WIDGET(filelist), -1, 0);
+#endif
+
   gtk_widget_set_name(filelist, "filelist");
 
   vadj = (GtkAdjustment *)gtk_adjustment_new(0, 0, 0, 0, 0, 0);
+#ifdef USEGTK2
+  gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(filelist),
+				GTK_ADJUSTMENT(vadj));
+#else
   gtk_clist_set_vadjustment(GTK_CLIST(filelist), GTK_ADJUSTMENT(vadj));
+#endif
   scrollbar = gtk_vscrollbar_new(GTK_ADJUSTMENT(vadj));
 
   if (mode_left_scroll) { 
@@ -700,12 +745,22 @@ int main(int argc, char **argv) {
   if (mode_start_mini) attribute_mini = 1;
 
   /* Attach the acceleration group to the window. */
+#ifdef USEGTK2
+  gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
+#else
   gtk_accel_group_attach(accel_group, GTK_OBJECT(window));
+#endif
 
   /* Set the window attributes and show it. */
+#ifdef USEGTK2
+  gtk_window_move(GTK_WINDOW(window), attribute_xpos, attribute_ypos);
+  if (attribute_width && attribute_height)
+    gtk_window_resize(GTK_WINDOW(window), attribute_width, attribute_height);
+#else
   gtk_widget_set_uposition(window, attribute_xpos, attribute_ypos);
   if (attribute_width && attribute_height)
     gtk_widget_set_usize(window, attribute_width, attribute_height);
+#endif
   if (attribute_mini) {
     attribute_mini = 1;
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(minicheckbox), 1);
