@@ -22,6 +22,9 @@
 #ifndef _XHIPPO_H
 #define _XHIPPO_H 1
 
+/* Pull in autoconf definitions. */
+#include "config.h"
+
 /* --- Other headers --- */
 
 #if HAVE_UNISTD_H
@@ -55,14 +58,21 @@
 #ifdef HAVE_LIBID3
 #include <id3.h>
 #endif
+#ifdef HAVE_LIBID3TAG
+#include <id3tag.h>
+#endif
 #ifdef HAVE_NFTW
 #define __USE_XOPEN_EXTENDED 1
 #include <ftw.h>
 #undef __USE_XOPEN_EXTENDED
 #endif
 #include <locale.h>
-#define ENABLE_NLS 1
-#include <libintl.h>
+
+/* This is necessary for MacOS X, where using the default redirection
+   mechanism generates binaries with unresolved symbols due to some
+   reasonably stupid kludges in gettext. */
+#define _INTL_REDIRECT_MACROS 1
+#include "gettext.h"
 #define _(X) gettext(X)
 
 
@@ -129,8 +139,9 @@ typedef enum _configtype configtype;
 struct _configcommand {
   gint pass;                /* Pass upon which to read this entry. */
   configtype type;          /* Type of this command. */
-  char *command;            /* Name of this command. */
+  const char *command;      /* Name of this command. */
   void *target;
+  const char *label;        /* Label for this command in the config dialog. */
 };
 typedef struct _configcommand configcommand;
 
@@ -151,7 +162,16 @@ extern GtkWidget *window, *restartbutton, *stopbutton, *buttons_box,
   *filelist, *statusbar, *nextbutton, *prevbutton, *pausebutton, *loadbutton,
   *minicheckbox, *randomcheckbox, *fileselector, *list_box,
   *scrollbar, *upbutton, *downbutton, *deletebutton, *popupmenu,
-  *up_item, *down_item, *delete_item, *info_item, *eventbox;
+  *up_item, *down_item, *delete_item, *info_item, *eventbox,
+  *prefs_window, *prefs_all_box;
+#ifdef USEGTK2
+extern GtkListStore *filestore;
+enum {
+  SONGINFO_COLUMN,
+  TITLE_COLUMN,
+  NUM_COLUMNS
+};
+#endif
 extern GtkAdjustment *vadj;
 extern GtkAccelGroup *accel_group;
 extern int sigchld_pipe[2];
@@ -213,7 +233,7 @@ void move_row(gint src, gint dest);
 void persist_playlist(void);
 gint filetype(char *file);
 void chomp(gchar *line);
-void status(char *message);
+void status(const char *message);
 void update_window(void);
 void stop_playing(void);
 void pause_playing(void);
@@ -234,62 +254,69 @@ int sort_compare_mtime(const void *a, const void *b);
 void sortplaylist(int (*func)(const void *a, const void *b),
 		  char needinfo);
 void read_playlist(gchar *name, int use_cwd);
-int write_playlist(gchar *name);
+int write_playlist(const gchar *name);
+void init_config_commands();
 void execute_command(gchar *command, gint pass);
 void read_config(gchar *name, gint pass);
 void read_config_files(gint pass);
-void handle_list_select(GtkWidget *gtklist, gint row, gint column,
+void show_prefs_window();
+#ifdef USEGTK2
+void handle_tree_select(GtkTreeSelection *selection, gpointer data);
+#else
+void handle_list_select(GtkCList *gtklist, gint row, gint column,
 			GdkEventButton *event, gpointer data);
-void handle_eventbox_click(GtkWidget *widget, GdkEventButton *event,
+#endif
+gboolean handle_eventbox_click(GtkWidget *widget, GdkEventButton *event,
+			       gpointer data);
+gboolean handle_menu_popup(GtkWidget *widget, GdkEventButton *event, 
 			   gpointer data);
-void handle_menu_popup(GtkWidget *widget, GdkEventButton *event, 
-		       gpointer data);
-void handle_fileselector_cancel(GtkWidget *widget, gpointer data);
-void handle_load_ok(GtkWidget *widget, gpointer data);
-void handle_save_ok(GtkWidget *widget, gpointer data);
-void handle_add_ok(GtkWidget *widget, gpointer data);
+void handle_fileselector_cancel(GtkButton *widget, gpointer data);
+void handle_load_ok(GtkButton *widget, gpointer data);
+void handle_save_ok(GtkButton *widget, gpointer data);
+void handle_add_ok(GtkButton *widget, gpointer data);
 #ifdef HAVE_NFTW
-void handle_add_dir_ok(GtkWidget *widget, gpointer data);
+void handle_add_dir_ok(GtkButton *widget, gpointer data);
 #endif
-void handle_next(GtkWidget *widget, gpointer data);
-void handle_prev(GtkWidget *widget, gpointer data);
+void handle_next(GtkButton *widget, gpointer data);
+void handle_prev(GtkButton *widget, gpointer data);
 void set_mini(void);
-void handle_minitoggle(GtkWidget *widget, gpointer data);
-void handle_randomtoggle(GtkWidget *widget, gpointer data);
-void handle_restart(GtkWidget *widget, gpointer data);
-void handle_stop(GtkWidget *widget, gpointer data);
-void handle_pause(GtkWidget *widget, gpointer data);
-gint handle_info_delete_event(GtkWidget *widget, GdkEvent *event, 
-			      gpointer data);
-void handle_info_destroy(GtkWidget *widget, gpointer data);
-void handle_menu_info(GtkWidget *widget, gpointer data);
-void handle_menu_up(GtkWidget *widget, gpointer data);
-void handle_menu_down(GtkWidget *widget, gpointer data);
-void handle_menu_delete(GtkWidget *widget, gpointer data);
-void handle_menu_user(gpointer data);
+void handle_minitoggle(GtkToggleButton *widget, gpointer data);
+void handle_randomtoggle(GtkToggleButton *widget, gpointer data);
+void handle_restart(GtkButton *widget, gpointer data);
+void handle_stop(GtkButton *widget, gpointer data);
+void handle_pause(GtkButton *widget, gpointer data);
+gboolean handle_info_destroy(GtkWidget *widget, GdkEvent *event, 
+			     gpointer data);
+void handle_menu_info(GtkMenuItem *widget, gpointer data);
+void handle_menu_up(GtkMenuItem *widget, gpointer data);
+void handle_menu_down(GtkMenuItem *widget, gpointer data);
+void handle_menu_delete(GtkMenuItem *widget, gpointer data);
+void handle_menu_user(GtkMenuItem *widget, gpointer data);
 void get_playlist_dir(gchar *buf, int buflen);
-void handle_menu_load(GtkWidget *widget, gpointer data);
-void handle_menu_save(GtkWidget *widget, gpointer data);
-void handle_menu_add(GtkWidget *widget, gpointer data);
+void handle_menu_load(GtkMenuItem *widget, gpointer data);
+void handle_menu_save(GtkMenuItem *widget, gpointer data);
+void handle_menu_add(GtkMenuItem *widget, gpointer data);
 #ifdef HAVE_NFTW
-void handle_menu_add_dir(GtkWidget *widget, gpointer data);
+void handle_menu_add_dir(GtkMenuItem *widget, gpointer data);
 #endif
-void handle_menu_sort_name(GtkWidget *widget, gpointer data);
-void handle_menu_sort_swapped_name(GtkWidget *widget, gpointer data);
-void handle_menu_sort_mtime(GtkWidget *widget, gpointer data);
-void handle_menu_clear(void);
-gint handle_delete_event(GtkWidget *widget, GdkEvent *event, gpointer data);
-void handle_destroy(GtkWidget *widget, gpointer data);
-void handle_configure(GtkWidget *widget, gpointer data);
+void handle_menu_sort_name(GtkMenuItem *widget, gpointer data);
+void handle_menu_sort_swapped_name(GtkMenuItem *widget, gpointer data);
+void handle_menu_sort_mtime(GtkMenuItem *widget, gpointer data);
+void handle_menu_clear(GtkMenuItem *widget, gpointer data);
+void handle_menu_prefs(GtkMenuItem *widget, gpointer data);
+void handle_menu_quit(GtkMenuItem *widget, gpointer data);
+gboolean handle_destroy(GtkWidget *widget, GdkEvent *event, gpointer data);
+gboolean handle_configure(GtkWidget *widget, GdkEventConfigure *event,
+			  gpointer data);
 void sigchld_backend(gpointer data, gint source, GdkInputCondition cond);
 void handle_sigchld(int dummy);
-#ifdef USEGNOMEMENUS
-void handle_about(GtkWidget *widget, void *data);
-#endif
 void handle_dnd_drop(GtkWidget *widget, GdkDragContext *context,
 		     gint x, gint y,
 		     GtkSelectionData *data, guint info,
-		     guint time);
+		     guint time, gpointer user_data);
+gint handle_prefs_delete_event(GtkWidget *widget, GdkEvent *event,
+			       gpointer data);
+void handle_prefs_destroy(GtkWidget *widget, gpointer data);
 void print_version(void);
 int main(int argc, char **argv);
 
